"""
Main API file.
"""

from typing import Union, Dict
import webbrowser

from fastapi import FastAPI, HTTPException, Query

from trainingsong.server.billboard_io import get_billboard_data
from trainingsong.server.spotify import (
    create_spotify_client,
    spotify_link,
    start_playback,
)
from trainingsong.server.db import get_tokens, database_session

app = FastAPI()


@app.get("/")
async def root(
    email: str,
    spotify_client_code: Union[str, None] = None,
    p: float = Query(..., ge=0, le=100),
    chart: str = "hot-100",
    autoplay: bool = False,
) -> Dict[str, Union[str, bool, float, None]]:
    """The main API endpoint. It takes in a percentage p, interacts with the billboard api and then redirects to the callback for the Spotify API."""

    print("Hit root endpoint")
    print(f"p: {p}")
    print(f"chart: {chart}")
    print(f"autoplay: {autoplay}")

    try:
        song_results = get_billboard_data(p, chart)
        song_results.autoplay = autoplay
    except HTTPException as e:
        raise HTTPException(status_code=404, detail=str(e)) from e

    if not spotify_client_code and not email:
        raise HTTPException(status_code=400, detail="Missing Spotify code and email")

    print("Got spotify code")

    song_info = song_results.song_info
    target_date = song_results.target_date

    try:
        if spotify_client_code is None and email is None:
            raise HTTPException(
                status_code=400, detail="Missing Spotify code and email"
            )
        sp = await create_spotify_client(spotify_client_code, email)
    except HTTPException as e:
        # raise HTTPException(
        #     status_code=404, detail=f"str(e). Failed to created Spotify client"
        # ) from e
        return {"errors": f"str(e). Failed to created Spotify client"}

    link, _name, uri = spotify_link(
        sp, song_results.song_name, song_results.artist_name
    )

    print(link)

    open_link = ""

    if autoplay:
        errors = attempt_play(sp, uri)
        if errors:
            errors += "Failed to start playback"
            open_link = "True"
    else:
        errors = ""
        open_link = "True"

    print(f"errors: {errors}")
    print(f"open_link: {open_link}")

    output = {
        "spotify_link": link,
        "song_name": song_results.song_name,
        "artist_name": song_results.artist_name,
        "target_date": target_date,
        "percentage": song_results.percentage,
        "chart": chart,
        "errors": errors,
        "song_info": song_info,
        "open_link": open_link,
    }

    return output


@app.get("/hello")
async def hello():
    return {"hello": "world"}


@app.get("/email_in_db")
async def email_in_db(email: str) -> Dict[str, bool]:
    async with database_session() as session:
        result = await get_tokens(email)
    return {"present_in_db": result is not None}


def attempt_play(sp, uri) -> str:
    "Attempt to play the song on Spotify"
    errors = ""
    devices = sp.devices()
    active_devices = devices["devices"] if devices else None

    if not active_devices:
        errors = "Unable to start playback because there are no active devices available. Please ensure that Spotify is active on one of your devices and try again."
    else:
        try:
            start_playback(sp, uri)

        except ValueError as e:
            errors = f"{str(e)}. Unable to start playback. Please ensure that Spotify is active on one of your devices and try again."
    return errors



"""Spotify API functions"""

from urllib.error import HTTPError
from typing import Tuple, Union, Optional
from dataclasses import dataclass
import os
import json
import time

import spotipy
from spotipy import SpotifyException
from spotipy.oauth2 import SpotifyOAuth
from fastapi import HTTPException
from trainingsong.server.db import (
    store_tokens,
    get_tokens,
    update_tokens,
    database_session,
)
from dotenv import load_dotenv

SCOPE = "user-modify-playback-state user-read-currently-playing user-read-recently-played user-read-playback-state"

# If running locally, load environment variables from .env
if os.environ.get("VERCEL") != "1":
    load_dotenv()


CLIENT_ID = os.environ.get("CLIENT_ID")
CLIENT_SECRET = os.environ.get("CLIENT_SECRET")

SPOTIFY_REDIRECT_URI = "http://localhost:8000/local_callback"


@dataclass
class StateData:
    """Dataclass for authenticate_spotify function"""

    song_name: str
    artist_name: str
    autoplay: Optional[bool]
    song_info: str
    target_date: str
    percentage: float
    chart: str


async def create_spotify_client(code: Union[str, None], email: str) -> spotipy.Spotify:
    """Create a Spotify client using the code from the Spotify API callback"""

    sp_oauth = SpotifyOAuth(
        client_id=CLIENT_ID,
        client_secret=CLIENT_SECRET,
        redirect_uri=SPOTIFY_REDIRECT_URI,
        scope=SCOPE,
    )

    async with database_session() as session:
        token_info = await get_tokens(email)

        if not token_info:
            print("Getting access token...")
            # Get the access token
            if code is None:
                raise ValueError("No code provided")
            try:
                token_info = sp_oauth.get_access_token(code)
            except:
                raise HTTPException(status_code=400, detail="Invalid Spotify code")
            if not token_info:
                raise HTTPException(status_code=400, detail="Invalid Spotify code")

            # Put token info into sqlalchemy database
            await store_tokens(
                email,
                token_info["access_token"],
                token_info["refresh_token"],
                token_info["expires_at"],
            )

        # If the access token is expired, refresh it
        if token_info["expires_at"] < time.time():
            print("Refreshing access token...")
            token_info = sp_oauth.refresh_access_token(token_info["refresh_token"])

            if token_info:
                # Put token info into sqlalchemy database
                await update_tokens(
                    email,
                    token_info["access_token"],
                    token_info["refresh_token"],
                    token_info["expires_at"],
                )
            else:
                raise ValueError("Failed to refresh access token. Please try again. ")

    access_token = token_info["access_token"] if token_info else None

    print("Got access token!")

    # Create a Spotify client with the access token
    sp = spotipy.Spotify(auth=access_token)
    print("Created Spotify client")

    return sp


def spotify_link(
    sp: spotipy.Spotify, song_name: str, artist_name: str
) -> Tuple[str, str, str]:
    """Get the Spotify link for the song using the Spotify API"""
    search_result = sp.search(q=f"{song_name} {artist_name}", type="track", limit=1)
    if search_result:
        song = search_result["tracks"]["items"][0]
    else:
        raise HTTPException(
            status_code=404,
            detail=f"Song {song_name} by {artist_name} not found on Spotify",
        )

    link = song["external_urls"]["spotify"]
    name = song["name"]
    uri = song["uri"]

    return link, name, uri


def start_playback(sp, uri, device_id=None) -> None:
    """Start playing the song on Spotify"""
    try:
        sp.start_playback(device_id=device_id, uris=[uri])
    except (HTTPError, SpotifyException) as e:
        ValueError(f"Spotify playback failed: {e}")



"""Entry point"""

from typing import Union, List, Optional, Tuple, Dict, Any, NamedTuple
import webbrowser
from threading import Thread
import time
import json
import os
import re

import requests
import uvicorn
from fastapi import FastAPI, Request

PROD_API = True

OAUTH_CODE = None
if PROD_API:
    URL = "https://training-song-api.vercel.app"
else:
    URL = "https://training-song-api-koayon.vercel.app"

AUTH_URL = "https://accounts.spotify.com/authorize?client_id=4259770654fb4353813dbf19d8b20608&response_type=code&redirect_uri=http%3A%2F%2Flocalhost%3A8000%2Flocal_callback&scope=user-modify-playback-state+user-read-currently-playing+user-read-recently-played+user-read-playback-state"
LOCAL_REDIRECT_URI = "http://localhost:8000/local_callback"


def _training_song(
    p: float,
    chart: Optional[str] = "hot-100",
    autoplay: Optional[bool] = True,
    verbose: Optional[bool] = True,
    email: Optional[str] = None,
) -> Tuple[Union[float, List[float], None], Dict[str, Any]]:
    """Return the training song for a given percentage
    Outputs: (acc, response)"""

    params = {
        "p": p,
        "autoplay": autoplay,
        "chart": chart,
        "spotify_client_code": OAUTH_CODE,
        "email": email,
    }

    raw_response = requests.get(
        URL,
        params=params,
        timeout=15,
    )

    response = raw_response.json()

    if verbose:
        print("Congrats your model got an accuracy of", p, "percent!")
        if response and "song_info" in response:
            print(response["song_info"])
        else:
            print("No song info found")
    if "errors" in response:
        print(response["errors"])

    if "open_link" in response and response["open_link"]:
        webbrowser.open(response["spotify_link"])

    return p, response


def ts(
    input_percentage: Union[float, List[float]],
    chart="hot-100",
    autoplay=True,
    verbose=True,
) -> Tuple[Union[float, List[float], None], Dict[str, Any]]:
    """Training song function.
    Starts a local server to capture the auth code from spotify and returns the song for your training accuracy.

    Args:
    input_percentage (float): The accuracy of your model.
    chart (str): The chart to use. Defaults to "hot-100".
    autoplay (bool): Whether to autoplay the song. Defaults to False.
    verbose (bool): Whether to print the song info. Defaults to False.

    Outputs:
    acc: The accuracy of your model. In the same form as you put it in.
    response: The response from the server as a dictionary.
    """

    email = _get_email()
    if not email:
        set_email()
        email = _get_email()

    if not email:
        raise ValueError(
            "No email found. Please run from the command line or add an email to a .email file in the root directory to proceed. "
        )

    email_in_db = _check_email(email)

    if not email_in_db:
        # start the local server in a new thread
        if not OAUTH_CODE:
            server_thread = Thread(target=_start_local_server)
            server_thread.start()

            # open the authorization URL in a browser
            webbrowser.open(AUTH_URL)

            # wait for the user to authorize and for the server to capture the OAuth code
            while not OAUTH_CODE:
                time.sleep(1)

    # if the input percentage is a list, we want to take the final value
    accuracy = (
        input_percentage
        if isinstance(input_percentage, (float, int))
        else input_percentage[-1]
    )

    # now we can call the training_song function with the captured OAuth code
    acc, response = _training_song(
        accuracy, chart=chart, autoplay=autoplay, verbose=verbose, email=email
    )
    return acc, response


local_app = FastAPI()


@local_app.get("/local_callback")
async def spotify_callback(request: Request):
    "Callback for the local server to capture the OAuth code"
    global OAUTH_CODE
    OAUTH_CODE = request.query_params.get("code")
    print("Got code:", OAUTH_CODE)
    return "Success! You can close this window."


def _start_local_server():
    "Start the local server"
    uvicorn.run(local_app, host="0.0.0.0", port=8000)


def _is_valid_email(email: str) -> bool:
    pattern = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"
    return bool(re.match(pattern, email))


def set_email():
    email_address = None
    while not email_address:
        potential_email = input("Enter your email address: ")
        if _is_valid_email(potential_email):
            email_address = potential_email
        else:
            print("Invalid email address. Please try again.")
    with open(".email", "w") as f:
        json.dump({"email": email_address}, f)


def _get_email():
    if not os.path.exists(".email"):
        return None
    with open(".email", "r") as f:
        try:
            email_dict = json.load(f)
        except json.decoder.JSONDecodeError:
            email_dict = {"email": None}
    return email_dict["email"]


def _check_email(email):
    response = requests.get(URL + "/email_in_db", params={"email": email})
    return response.json()["present_in_db"]


if __name__ == "__main__":
    INPUT_PERCENTAGE = None
    RAW_INPUT = ""
    while not INPUT_PERCENTAGE:
        RAW_INPUT = input("How well did your model do? (Enter a percentage): ")
        if RAW_INPUT.isdigit():
            INPUT_PERCENTAGE = float(RAW_INPUT)

    ts(INPUT_PERCENTAGE, verbose=True, autoplay=True)



"Billboard API calls and data processing"

import datetime
from typing import Tuple
from dataclasses import dataclass

from fastapi import HTTPException
import billboard

from trainingsong.server.spotify import StateData


@dataclass
class Song:
    """A dataclass to store the song name, artist name and number of weeks on the chart"""

    artist: str
    weeks: int
    title: str


def get_billboard_data(
    percentage: float,
    chart: str = "hot-100",
) -> StateData:
    """Call Billboard API and get the song name, artist name and song info"""
    if percentage > 100 or percentage < 0:
        raise ValueError("Please enter a percentage between 0 and 100")
    if percentage < 1:
        # Turn a decimal into a percentage
        percentage *= 100

    try:
        number_one_song, target_date = get_number_one_song(percentage, chart)
    except HTTPException:
        raise HTTPException(status_code=404, detail="No chart data found")

    song_name = number_one_song.title
    artist_name = number_one_song.artist

    song_info = f"""The Number 1 song {percentage}% through the 1900s on the {chart} chart was {song_name} by {artist_name}. \nThe date was {target_date} and the song was on the chart for {number_one_song.weeks} weeks."""
    result = StateData(
        song_name=song_name,
        artist_name=artist_name,
        autoplay=None,
        song_info=song_info,
        target_date=str(target_date),
        percentage=percentage,
        chart=chart,
    )
    return result


def get_number_one_song(
    percentage: float, chart: str = "hot-100"
) -> Tuple[Song, datetime.date]:
    """Get the number one song on the chosen Billboard chart on date that is percentage% through the 1900s"""
    target_year = int(percentage)

    # Calculate the target date based on the fractional part of the percentage
    fractional_percentage = percentage % 1
    days_in_year = (
        366
        if target_year % 4 == 0 and (target_year % 100 != 0 or target_year % 400 == 0)
        else 365
    )
    target_day = int(days_in_year * fractional_percentage)

    target_date = datetime.date(1900 + target_year, 1, 1) + datetime.timedelta(
        days=target_day
    )

    # Fetch the Billboard Hot 100 chart data for the target date
    chart_output = billboard.ChartData(chart, date=target_date)

    # Get the Number 1 song on the chart
    if chart_output:
        number_one_song = chart_output[0]
    else:
        raise HTTPException(status_code=404, detail="No chart data found")

    return number_one_song, target_date



"""Database module for storing and retrieving user tokens from database."""

import databases
import sqlalchemy
from sqlalchemy import Table, Column, String, text, Integer
import os
from dotenv import load_dotenv
from typing import Optional, Dict
import asyncio
from contextlib import asynccontextmanager

# If running locally, load environment variables from .env
if os.environ.get("VERCEL") != "1":
    load_dotenv()

DATABASE_URL = os.environ.get("DATABASE_URL")
if DATABASE_URL is None:
    raise ValueError("DATABASE_URL environment variable not set or empty")

database = databases.Database(DATABASE_URL)
metadata = sqlalchemy.MetaData()

tokens = Table(
    "tokens",
    metadata,
    Column("email", String, primary_key=True, unique=True),
    Column("access_token", String),
    Column("refresh_token", String),
    Column("expires_at", Integer),
)

engine = sqlalchemy.create_engine(DATABASE_URL)


def create() -> None:
    with engine.begin() as connection:
        connection.execute(text("DROP TABLE IF EXISTS tokens"))
    metadata.create_all(engine)


async def store_tokens(
    email: str, access_token: str, refresh_token: str, expires_at: int
) -> None:
    query = tokens.insert().values(
        email=email,
        access_token=access_token,
        refresh_token=refresh_token,
        expires_at=expires_at,
    )
    await database.execute(query)


async def get_tokens(email: str):
    # Return type is dictionary like. It's a Record object from databases
    query = tokens.select().where(tokens.c.email == email)
    result = await database.fetch_one(query)
    return result


async def delete_tokens(email: str) -> None:
    query = tokens.delete().where(tokens.c.email == email)
    await database.execute(query)


async def update_tokens(
    email: str, access_token: str, refresh_token: str, expires_at: str
) -> None:
    query = (
        tokens.update()
        .where(tokens.c.email == email)
        .values(
            access_token=access_token,
            refresh_token=refresh_token,
            expires_at=expires_at,
        )
    )
    await database.execute(query)


@asynccontextmanager
async def database_session():
    await database.connect()
    try:
        yield
    finally:
        await database.disconnect()


async def main():
    async with database_session():
        # Test store_tokens and get_tokens
        # create()

        # await store_tokens("test", "test", "test", 1689727126)
        # record = await get_tokens("test")
        # if record:
        #     print(record)
        #     print(record.email)
        pass


if __name__ == "__main__":
    asyncio.run(main())
